package PktDump::Analyser;

use strict;
use warnings;

use Module::Pluggable
   search_path => [ "PktDump::Analyser" ];

my @analysers = plugins;

use Attribute::Storage 0.06 qw( get_subattr );

sub Filter_description :ATTR(CODE,RAWDATA)
{
   my $pkg = shift;
   my ( $desc ) = @_;
   return $desc;
}

sub Filter_option :ATTR(CODE,MULTI)
{
   my $pkg = shift;
   my ( $opts, $name, $type ) = @_;

   $opts and exists $opts->{$name} and
      die "Already have an option called '$name'\n";

   my $code = $pkg->can( "gen_opttype_$type" ) or
      die "No option type called $type\n";

   $opts->{$name} = $code->( $pkg, name => $name );

   return $opts;
}

sub Filter_argument :ATTR(CODE,MULTI)
{
   my $pkg = shift;
   my ( $args, $name, $type ) = @_;

   my $code = $pkg->can( "gen_opttype_$type" ) or
      die "No option type called $type\n";

   push @$args, $code->( $pkg, name => $name );

   return $args;
}

# Some common filter option types
my $match_number = qr/0x[[:xdigit:]]+|\d+/;

sub gen_opttype_string { shift; PktDump::OptType->new( @_ ) }

sub gen_opttype_number { shift; PktDump::OptType::number->new( @_ ) }
{
   package PktDump::OptType::number;
   use base qw( PktDump::OptType );

   sub match { $match_number }
   sub parse { m/^0x/ ? hex : $_ }
}

sub gen_opttype_numbers { shift; PktDump::OptType::numbers->new( @_ ) }
{
   package PktDump::OptType::numbers;
   use base qw( PktDump::OptType );

   sub match { qr/$match_number(?:,$match_number)*/o }
   sub parse { [ map m/^0x/ ? hex : $_, split m/,/, $_ ] }
}

sub gen_opttype_dirnumbers { shift; PktDump::OptType::dirnumbers->new( @_ ) }
{
   package PktDump::OptType::dirnumbers;
   use base qw( PktDump::OptType::Directional PktDump::OptType::numbers );
}

sub gen_opttype_flag { shift; PktDump::OptType::flag->new( @_ ) }
{
   package PktDump::OptType::flag;
   use base qw( PktDump::OptType );

   sub desc_arg { shift->{name} }
   sub extract_value { 1 }
}

my %type2a;
sub find
{
   shift;
   my ( $type, $key ) = @_;

   return $type2a{$type}{$key} if $type2a{$type}{$key};
   
   foreach my $class ( @analysers ) {
      next unless $class->can( $type ) and $class->$type == $key;

      my $a = $class->new( $type => $key );
      return $type2a{$type}{$key} = $a;
   }

   die sprintf "Unable to find an Analyser for %s => %d (0x%x)\n", $type, $key, $key;
}

sub for_hatype    { shift->find( HATYPE    => shift ) }
sub for_ethertype { shift->find( ETHERTYPE => shift ) }
sub for_ipproto   { shift->find( IPPROTO   => shift ) }

sub new
{
   my $class = shift;
   return bless {}, $class;
}

sub format_hex
{
   my $self = shift;
   my ( $data, $step ) = @_;

   $step ||= 16;

   my $offs = 0;
   my @lines;
   while( length $data ) {
      ( my $chunk, $data ) = unpack( "a$step a*", $data );
      my $fill = $step - length $chunk;
      push @lines, sprintf( "  [%04x] %s | %s", $offs,
         join( " ", map( sprintf("%02x",$_), unpack( "(C)$step", $chunk ) ), ("  ") x $fill ),
         join( "", map length($_)?(m/[\x20-\x7e]/?$_:"."):" ", unpack( "(a)$step", $chunk ) ),
      );
      $offs += $step;
   }

   @lines;
}

sub format_protaddr
{
   my $self = shift;
   my ( $ethertype, $addr, $opts ) = @_;

   if( my $a = eval { $self->for_ethertype( $ethertype ) } ) {
      $a->format_this_protaddr( $addr, $opts );
   }
   else {
      sprintf "[%04x]%s", $ethertype, join ":", map sprintf("%02x",$_), unpack "C*", $addr;
   }
}

sub format_hwaddr
{
   my $self = shift;
   my ( $hwtype, $addr, $opts ) = @_;

   if( my $a = eval { $self->for_hatype( $hwtype ) } ) {
      $a->format_this_hwaddr( $addr, $opts );
   }
   else {
      sprintf "[%04x]%s", $hwtype, join ":", map sprintf("%02x",$_), unpack "C*", $addr;
   }
}

sub gen_filter
{
   my $self = shift;
   my ( $text ) = @_;

   my $filter = $self->parse_filter_any( $text );
   length $text == 0 or die "Expected end of filter; found >> $text\n";

   return $filter->compile;
}

our $DEBUG_INDENT = "";

sub parse_filter_any
{
   my $self = shift;

   print STDERR "${DEBUG_INDENT}parse_filter_any >> $_[0]\n" if $ENV{DEBUG_PKTDUMP_FILTER};

   my @filters = $self->parse_filter_all( $_[0] );

   while( length $_[0] ) {
      $_[0] =~ s/^or\s*//;

      my $t = $_[0];
      eval { push @filters, $self->parse_filter_all( $t ) } or last;
      $_[0] = $t;
   }

   return $filters[0] if @filters == 1;
   return PktDump::Filter->any( @filters );
}

sub parse_filter_all
{
   my $self = shift;

   print STDERR "${DEBUG_INDENT}parse_filter_all >> $_[0]\n" if $ENV{DEBUG_PKTDUMP_FILTER};

   my @filters = $self->parse_filter_not( $_[0] );

   while( length $_[0] ) {
      $_[0] =~ s/^(and)?\s*//;

      my $t = $_[0];
      eval { push @filters, $self->parse_filter_not( $t ) } or last;
      $_[0] = $t;
   }

   return $filters[0] if @filters == 1;
   return PktDump::Filter->all( @filters );
}

sub parse_filter_not
{
   my $self = shift;

   print STDERR "${DEBUG_INDENT}parse_filter_not >> $_[0]\n" if $ENV{DEBUG_PKTDUMP_FILTER};

   return $self->parse_filter( $_[0] ) unless $_[0] =~ s/^not\b\s*//;

   my $filter = $self->parse_filter( $_[0] );
   return PktDump::Filter->not( $filter );
}

sub parse_filter
{
   my $self = shift;

   print STDERR "${DEBUG_INDENT}parse_filter >> $_[0]\n" if $ENV{DEBUG_PKTDUMP_FILTER};

   $_[0] =~ s/^\s+//;

   if( $_[0] =~ s/^\(\s*// ) {
      local $DEBUG_INDENT = "$DEBUG_INDENT  ";
      my $filter = $self->parse_filter_any( $_[0] );
      $_[0] =~ s/^\)\s*// or die "Expected ); found >> $_[0]\n";

      return $filter;
   }

   $_[0] =~ s/^(\w+)\s*// or die "Expected a keyword; found >> $_[0]\n";
   my $kw = $1;

   my $method = "handle_filter_$kw";

   my $mref = $self->can( $method );

   if( !$mref ) {
      foreach my $class ( @analysers ) {
         next unless $mref = $class->can( $method );
         last;
      }
   }

   die "Unable to handle filter keyword '$kw'\n" unless $mref;

   my $argspecs = get_subattr( $mref, "Filter_argument" );
   my @args;

   foreach my $argspec ( @$argspecs ) {
      defined( my $value = $argspec->extract_value( $_[0] ) ) or
         die "Expected value for argument $argspec->{name}\n";

      push @args, $value;
   }

   my $optspecs = get_subattr( $mref, "Filter_option" );
   my %opts;

   OPT: while( keys %$optspecs and length $_[0] ) {
      foreach my $opt ( keys %$optspecs ) {
         my $optspec = $optspecs->{$opt};

         defined( my $value = $optspec->extract_option( $_[0] ) ) or next;

         # Turn spaces into underscores so it's easier for handler code
         $opt =~ s/ /_/g;
         $opts{$opt} = $value;

         next OPT;
      }

      last;
   }

   return $self->$mref( @args, \%opts );
}

# Some helpers for subclasses

sub build_dir_filter
{
   my $self = shift;
   my ( $dir, $srcfilter, $dstfilter ) = @_;

   if( !defined $dir ) {
      return PktDump::Filter->any( $srcfilter, $dstfilter );
   }
   elsif( $dir eq "src" ) {
      return $srcfilter;
   }
   elsif( $dir eq "dst" ) {
      return $dstfilter;
   }
   elsif( $dir eq "local" ) {
      return PktDump::Filter->any(
         PktDump::Filter->all( $self->handle_filter_rx, $dstfilter ),
         PktDump::Filter->all( $self->handle_filter_tx, $srcfilter ),
      );
   }
   elsif( $dir eq "remote" ) {
      return PktDump::Filter->any(
         PktDump::Filter->all( $self->handle_filter_rx, $srcfilter ),
         PktDump::Filter->all( $self->handle_filter_tx, $dstfilter ),
      );
   }
   else {
      die "Unrecognised direction $dir\n";
   }
}

sub build_layer_filter
{
   my $self = shift;
   my ( $key, $value ) = @_;

   my @matchers;

   my $base_mref = $self->can( "handle_layer_filter_$key" );
   if( $base_mref ) {
      push @matchers, $base_mref->( $self, $value );
   }

   foreach my $analyser ( @analysers ) {
      next unless my $mref = $analyser->can( "handle_layer_filter_$key" );
      next if $base_mref and $mref == $base_mref;

      push @matchers, $mref->( $analyser, $value );
   }

   @matchers or die "Could not find any Analysers to build a layer filter for $key\n";

   my $layer = PktDump::Filter->layer;
   foreach my $m ( @matchers ) {
      $layer->matcher( @$m );
   }

   return $layer;
}

# Now the metadata filters

use PktDump::Filter;

use Linux::SocketFilter qw( SKF_AD_OFF SKF_AD_PKTTYPE SKF_AD_IFINDEX SKF_AD_PROTOCOL );
use Socket::Packet qw( PACKET_OUTGOING );

sub handle_filter_rx
   :Filter_description(Received packets)
{
   my $self = shift;
   return PktDump::Filter->not( $self->handle_filter_tx );
}

sub handle_filter_tx
   :Filter_description(Transmitted packets)
{
   return PktDump::Filter->match_u32( SKF_AD_OFF + SKF_AD_PKTTYPE, PACKET_OUTGOING );
}

sub handle_filter_ifindex
   :Filter_description(Packets on the given interface index)
   :Filter_argument(index => 'numbers')
{
   my $self = shift;
   my ( $index ) = @_;

   return PktDump::Filter->match_u32( SKF_AD_OFF + SKF_AD_IFINDEX, $index );
}

sub handle_filter_ifname
   :Filter_description(Packets on the given interface)
   :Filter_argument(name => 'string')
{
   my $self = shift;
   my ( $name ) = @_;

   defined( my $index = IO::Socket::Packet->ifname2index( $name ) ) or
      die "No interface called '$name'";

   return $self->handle_filter_ifindex( $index );
}

sub handle_filter_proto
   :Filter_description(Packets of the given protocol type)
   :Filter_argument(proto => 'numbers')
{
   my $self = shift;
   my ( $proto ) = @_;

   return PktDump::Filter->match_u32( SKF_AD_OFF + SKF_AD_PROTOCOL, $proto );
}

sub handle_layer_filter_proto
{
   my $self = shift;
   my ( $proto ) = @_;
   
   return [ $self->handle_filter_proto( $proto ),
      offset => "NET"
   ];
}

# Now actually load the plugins
foreach ( @analysers ) {
   ( my $f = "$_.pm" ) =~ s{::}{/}g;
   require $f;
}

1;
