package PktDump::Analyser::Ether802dot3;

use strict;
use warnings;

use base qw( PktDump::Analyser::Ethernet );

use constant HATYPE => 1;

use PktDump::Filter;

sub analyse
{
   my $self = shift;
   my ( $pkt, $opts ) = @_;

   my ( $dst, $src, $ethertype, $data ) = unpack( "a6 a6 n a*", $pkt );

   $_ = $self->format_mac( $_, $opts ) for $src, $dst;

   my $linkline = $opts->{one_line} ? sprintf "{%s->%s}", $src, $dst
                                    : sprintf "802.3| %s->%s: type=%04x", $src, $dst, $ethertype;

   return $self->analyse_ethernet( $linkline, $ethertype, $data, $opts );
}

sub hdrlen
{
   my $self = shift;
   my ( $info, $packet ) = @_;

   # Account for 802.1q VLAN
   if( $info->{proto} == PktDump::Analyser::Ether802dot1q->ETHERTYPE ) {
      return 14 + 4;
   }
   else {
      return 14;
   }
}

*format_this_hwaddr = __PACKAGE__->can( "format_mac" );

sub gen_opttype_macaddr { shift; PktDump::Analyser::Ether802dot3::MacAddrOptType->new( @_ ) }
{
   package PktDump::Analyser::Ether802dot3::MacAddrOptType;
   use base qw( PktDump::OptType::Directional PktDump::OptType );

   my $octet = qr/[0-9a-f]{2}/i;
   sub match { qr/(?:$octet:){5}$octet/ }
   sub parse { [ map hex, split m/:/, $_ ] }
}

sub handle_filter_ether
   :Filter_description(Ethernet 802.3)
   :Filter_option(type => 'numbers')
   :Filter_option(addr => 'macaddr')
   :Filter_option(broadcast => 'flag')
   :Filter_option(unicast => 'flag')
{
   my $self = shift;
   my ( $opts ) = @_;

   if( $opts->{broadcast} and $opts->{unicast} ) {
      die "Cannot specify ether broadcast and ether unicast\n";
   }

   my @parts;

   # TODO: One day, we'll put a match on SKF_AD_HATYPE but that isn't yet in
   # mainline kernel

   if( my $type = $opts->{type} ) {
      push @parts, PktDump::Filter->match_u16( 12, $type );
   }
   if( my $addr = $opts->{addr} ) {
      my ( $dir, $octets ) = @$addr;
      my @o = @$octets;

      push @parts, $self->build_dir_filter( $dir,
         # Arrange these to optimise 32bit boundaries
         PktDump::Filter->all( # src
            PktDump::Filter->match_u16( 6, $o[1]|$o[0]<<8 ),
            PktDump::Filter->match_u32( 8, $o[5]|$o[4]<<8|$o[3]<<16|$o[2]<<24 ),
         ),
         PktDump::Filter->all( # dst
            PktDump::Filter->match_u32( 0, $o[3]|$o[2]<<8|$o[1]<<16|$o[0]<<24 ),
            PktDump::Filter->match_u16( 4, $o[5]|$o[4]<<8 ),
         ),
      );
   }
   if( $opts->{broadcast} ) {
      # broadcast == ether dst addr ff:ff:ff:ff:ff:ff
      push @parts,
         PktDump::Filter->match_u32( 0, 0xffffffff ),
         PktDump::Filter->match_u16( 4, 0xffff );
   }
   if( $opts->{unicast} ) {
      # unicast == not( ether dst addr ff:ff:ff:ff:ff:ff )
      push @parts, PktDump::Filter->not( PktDump::Filter->all(
         PktDump::Filter->match_u32( 0, 0xffffffff ),
         PktDump::Filter->match_u16( 4, 0xffff ),
      ) );
   }

   PktDump::Filter->all( @parts );
}

1;
