package PktDump::Analyser::ICMP;

use strict;
use warnings;

use base qw( PktDump::Analyser );

use constant IPPROTO => 1;

# See also
#   http://www.faqs.org/rfcs/rfc792.html

sub analyse
{
   my $self = shift;
   my ( $pkt, $opts, $outer ) = @_;

   my ( $type, $code, $cksum, $data ) = unpack(
       "C      C      n       a*", $pkt );

   my $hosts = "";
   if( $opts->{one_line} ) {
      $hosts = "$outer->{src}->$outer->{dst} ";
   }

   if( my $mth = $self->can( "analyse_icmp_$type" ) ) {
      "ICMP| ".$hosts . $mth->( $self, $opts, $code, $data );
   }
   else {
      sprintf "ICMP| %stype=%d code=%d", $hosts, $type, $code;
   }
}

sub analyse_icmp_0 { &analyse_icmp_echo( shift, "echo-reply", @_ ) }
sub analyse_icmp_8 { &analyse_icmp_echo( shift, "echo-request", @_ ) }
sub analyse_icmp_echo
{
   my $self = shift;
   my ( $name, $opts, $code, $data ) = @_;
   my ( $id, $seq ) = unpack( "n n", $data );
   return "$name seq=$seq" . ( $opts->{verbose} ? " id=$id" : "" );
}

my %dst_unreachable = (
   0 => "net-unreachable",
   1 => "host-unreachable",
   2 => "proto-unreachable",
   3 => "port-unreachable",
   4 => "frag-needed",
   5 => "src-route-failed",
);
sub analyse_icmp_3
{
   my $self = shift;
   my ( $opts, $code, $data ) = @_;
   return $dst_unreachable{$code} || "dst-unreachable code=$code";
}

sub analyse_icmp_4
{
   my $self = shift;
   return "source-quench";
}

my %redirect = (
   0 => "network-redirect",
   1 => "host-redirect",
   2 => "tos-network-redirect",
   3 => "tos-host-redirect",
);
sub analyse_icmp_5
{
   my $self = shift;
   my ( $opts, $code, $data ) = @_;
   my ( $gw ) = unpack( "a4", $data );
   $gw = PktDump::Analyser::IPv4->format_ip( $gw, $opts );
   return ( $redirect{$code} || "redirect code=$code" ) . " gw=$gw";
}

my %time_exceeded = (
   0 => "ttl-exceeded",
   1 => "frag-reassembly-time-exceeded",
);
sub analyse_icmp_11
{
   my $self = shift;
   my ( $opts, $code ) = @_;
   return $time_exceeded{$code} || "time-exceeded code=$code";
}

sub analyse_icmp_12
{
   my $self = shift;
   my ( $opts, $code, $data ) = @_;
   my ( $ptr ) = unpack( "C", $data );
   return "parameter-problem ptr=$ptr";
}

sub analyse_icmp_13 { &analyse_icmp_timestamp( shift, "timestamp", @_ ) }
sub analyse_icmp_14 { &analyse_icmp_timestamp( shift, "timestamp-reply", @_ ) }
sub analyse_icmp_timestamp
{
   my $self = shift;
   my ( $name, $opts, $code, $data ) = @_;
   my ( $id, $seq, $origtime, $rxtime, $txtime ) = unpack( "n n N N N", $data );
   $_ = sprintf "%02d:%02d:%02d.%03d", $_/3600000, ($_/60000)%60, ($_/1000)%60, $_%1000
      for $origtime, $rxtime, $txtime;
   return "$name orig=$origtime rx=$rxtime tx=$txtime seq=$seq" . ( $opts->{verbose} ? " id=$id" : "" );
}

sub handle_filter_icmp
   :Filter_description(IP Control Message Protocol)
   :Filter_option(type => 'numbers')
   :Filter_option(code => 'numbers')
{
   my $self = shift;
   my ( $opts ) = @_;

   my @parts = my $below = $self->build_layer_filter( ipproto => IPPROTO );

   if( defined( my $type = $opts->{type} ) ) {
      push @parts, $below->match_u8( 0, $type );
   }
   if( defined( my $code = $opts->{code} ) ) {
      push @parts, $below->match_u8( 1, $code );
   }

   return PktDump::Filter->all( @parts );
}

1;
