package PktDump::Analyser::IPv4;

use strict;
use warnings;

use base qw( PktDump::Analyser );

use constant ETHERTYPE => 0x0800;
use constant PPPDLL_PROTO => 0x0021;

use PktDump::Filter;

use Socket qw( inet_ntoa pack_sockaddr_in );
use Socket::GetAddrInfo qw( :newapi getnameinfo );

sub analyse
{
   my $self = shift;
   my ( $pkt, $opts ) = @_;

   my ( $vl ) = unpack( "C", $pkt );
   ( $vl & 0xf0 ) == 0x40 or die "Expected 4\n";
   my $hlen = ( $vl & 0x0f ) * 4;

   my ( $hdr, $data ) = unpack( "a$hlen a*", $pkt );
   my ( $ds, $len, $id, $flagfrag, $ttl, $prot, $cksum, $src, $dst, $ipopts ) = unpack(
     "x C    n     n    n          C     C      n       a4    a4    a*", $hdr );
   my $frag  = $flagfrag & 0x1fff;

   $_ = $self->format_ip( $_, $opts ) for $src, $dst;

   my $line = sprintf "IPv4| %s->%s: proto=%d", $src, $dst, $prot;

   $line .= " ttl=$ttl" if $opts->{verbose};

   $line .= " DF" if $flagfrag & 0x4000;
   $line .= " MF" if $flagfrag & 0x2000;

   $line .= " offs=$frag" if $frag > 0;

   $line .= " id=$id" if $frag > 0 or $flagfrag & 0x2000;

   return $line if $frag > 0;

   my @lines;
   push @lines, $line unless $opts->{one_line};

   return @lines, $self->for_ipproto( $prot )->analyse( $data, $opts, { src => $src, dst => $dst } );
}

my %namecache;
sub format_ip
{
   my $self = shift;
   my ( $addr, $opts ) = @_;

   if( $opts->{resolve_hosts} ) {
      return $namecache{$addr} if exists $namecache{$addr};

      my ( $err, $name ) = getnameinfo( pack_sockaddr_in( 0, $addr ), 0 );
      return $namecache{$addr} = $name;
   }

   return inet_ntoa( $addr );
}

*format_this_protaddr = \&format_ip;

sub gen_opttype_addrmask { shift; PktDump::Analyser::IPv4::AddrMaskOptType->new( @_ ) }
{
   package PktDump::Analyser::IPv4::AddrMaskOptType;
   use base qw( PktDump::OptType::Directional PktDump::OptType );

   use Socket qw( inet_aton );

   use constant match => qr{[^ /]+(?:/\d+)?};
   use constant desc  => q{$ADDR[/$SUBNET]};
   sub parse {
      my ( $host, $prefix ) = m{^([^ /]+)(?:/(\d+))?$};
      defined $prefix or $prefix = 32;

      $prefix <= 32 or die "Prefix length must be 0-32\n";

      my $hostnum = inet_aton( $host );
      defined $hostnum or die "Unknown host '$host'\n";

      $hostnum = unpack( "N", $hostnum );

      return [ $hostnum ] if $prefix == 32;

      my $netmask = 0xffffffff ^ ( ( 1 << ( 32 - $prefix ) ) - 1 );
      $hostnum &= $netmask;

      return [ $hostnum, mask => $netmask ];
   }
}

sub handle_layer_filter_ipproto
{
   my $self = shift;
   my ( $proto ) = @_;

   my $below = $self->build_layer_filter( proto => ETHERTYPE );

   return $below->extend(
      $below->match_u8( 9, $proto ),

      genoffset => sub {
         $below->loadA_indir( "BYTE", 0 ),
         "AND 15",
         "LSH 2",
      },
      offset => "A",
   );
}

sub handle_filter_ip
   :Filter_description(Internet Protocol v4 (IPv4))
   :Filter_option(proto => 'numbers')
   :Filter_option(host  => 'addrmask')
{
   my $self = shift;
   my ( $opts ) = @_;

   my $below = $self->build_layer_filter( proto => ETHERTYPE );

   my @parts = $below;
   
   if( my $proto = $opts->{proto} ) {
      push @parts, $below->match_u8( 9, $proto );
   }
   if( my $host = $opts->{host} ) {
      my ( $dir, $hostspec ) = @$host;
      push @parts, $self->build_dir_filter( $dir,
         $below->match_u32( 12, @$hostspec ), # src
         $below->match_u32( 16, @$hostspec ), # dst
      );
   }

   return PktDump::Filter->all( @parts );
}

1;
