package PktDump::Capfile::PcapNg;

# See also
#   http://www.winpcap.org/ntar/draft/PCAP-DumpFileFormat.html

use strict;
use warnings;

use base qw( PktDump::Capfile );
use constant FORMAT => "pcapng";

use Socket::Packet qw(
   PACKET_HOST PACKET_BROADCAST PACKET_MULTICAST PACKET_OTHERHOST PACKET_OUTGOING
);

sub write_block
{
   my $self = shift;
   my ( $type, $content ) = @_;

   my $len = (3 + length $content) & ~3; # round up to multiple of 4

   $len += 12; # account for type/len*2

   $self->pack( "i i a* x![i] i", $type, $len, $content, $len );
}

sub pack_block
{
   my $self = shift;
   my ( $type, $fmt, @args ) = @_;

   $self->write_block( $type, pack( $fmt, @args ) );
}

sub _pack_options
{
   my $self = shift;

   my $ret = "";
   while( @_ ) {
      my $type = shift;
      my $val = shift;

      $ret .= pack( "s s a* x![i]", $type, length $val, $val );
   }

   $ret .= pack( "s s", 0, 0 ); # opt_endofopt

   return $ret;
}

sub write_header
{
   my $self = shift;

   $self->pack_block( 0x0A0D0D0A,
      "i s s q",
      0x1A2B3C4D, # byte-order magic
      1, 0,       # maj,min vers
      -1,         # total length - unknown as yet
   );
}

my %hatype2linktype = (
   1, 1, # Ethernet
);

sub write_iface_block
{
   my $self = shift;
   my ( $info ) = @_;

   my $linktype = $hatype2linktype{$info->{hatype}};
   defined $linktype or
      die "TODO: Not sure what pcap linktype for $info->{hatype}\n";

   $self->pack_block( 0x00000001,
      "s x[s] i a*",
      $linktype, 65536,

      $self->_pack_options( 
         2 => $info->{ifname}, # if_name
         9 => pack( "C", 6 ),  # if_tsresol == 10^-6 seconds
         # TODO: consider other options; IPv4/6 addr, MAC addr, ...
      ),
   );

   return $self->{next_interface_id}++ || 0;
}

sub write_packet
{
   my $self = shift;
   my ( $info, $packet ) = @_;

   # pcapng file numbers interfaces implictly by ordering of the interface
   # definition blocks

   my $interface_id;

   unless( defined( $interface_id = $self->{ifindex2id}{ $info->{ifindex} } ) ) {
      $interface_id = $self->write_iface_block( $info );

      $self->{ifindex2id}{ $info->{ifindex} } = $interface_id;
   }

   my $timestamp = $info->{timestamp} * 1000000; # units of 10^-6 seconds

   my $pkttype = $info->{pkttype};
   my $flags = 
      ( $pkttype == PACKET_OUTGOING ? 2 : 1 ) |

      ( $pkttype == PACKET_HOST      ? 1 :
        $pkttype == PACKET_MULTICAST ? 2 :
        $pkttype == PACKET_BROADCAST ? 3 :
        $pkttype == PACKET_OTHERHOST ? 4 : 0 ) << 2;

   $self->pack_block( 0x00000006,
      "i i i i i a* x![i] a*",
      $interface_id,
      $timestamp >> 32, $timestamp & 0xffffffff,
      length $packet, length $packet,
      $packet,

      $self->_pack_options(
         2 => $flags, # ebp_flags
      ),
   );
}

1;
