package PktDump::Filter;

use strict;
use warnings;

use Linux::SocketFilter::Assembler qw( assemble );

use Carp;

# Private constructor; use the API
sub _new
{
   my $class = shift;
   bless {}, $class;
}

sub instr
{
   my $self = shift;
   push @{ $self->{instr} }, @_;
   return $self;
}

sub instructions
{
   my $self = shift;
   return @{ $self->{instr} };
}

sub loadA_indir
{
   my $self = shift;
   my ( $size, $offset ) = @_;

   $offset =~ m/^-?\d+$/ or die "Expected immediate offset; found $offset\n";

   return "LD $size\[$offset]";
}

sub compile
{
   my $self = shift;

   my $ret = "";

   my $accept = PktDump::Filter->_new->
      instr( "LD len" )->
      instr( "RET A" );

   my $program = PktDump::Filter->all( $self, $accept )->
      instr( "RET 0" );

   # Need to pass -1 as $jf, to ensure it hits the RET 0 part
   my @instr = $program->instructions( 0, -1 );

   if( $ENV{DEBUG_PKTDUMP_FILTER} ) {
      print STDERR "> $_\n" for @instr;
   }

   return assemble( join "\n", @instr );
}

# Public API

sub match_u8  { my $self = shift; PktDump::Filter::Match->_new( "BYTE", @_, loader => $self ); }
sub match_u16 { my $self = shift; PktDump::Filter::Match->_new( "HALF", @_, loader => $self ); }
sub match_u32 { my $self = shift; PktDump::Filter::Match->_new( "WORD", @_, loader => $self ); }

sub all
{
   my $class = shift;
   my @parts = @_;

   die "TODO: All of nothing?\n" unless @parts;

   my $f = PktDump::Filter::All->_new;
   $f->_part( $_ ) for @parts;
   return $f;
}

sub any
{
   my $class = shift;
   my @parts = @_;

   die "TODO: Any of nothing?\n" unless @parts;

   my $f = PktDump::Filter::Any->_new;
   $f->_part( $_ ) for @parts;
   return $f;
}

sub not
{
   my $self = shift;
   return PktDump::Filter::Not->_new( shift );
}

sub layer
{
   my $self = shift;
   return PktDump::Filter::Layer->_new;
}

package PktDump::Filter::Match;
use base qw( PktDump::Filter );

use Carp;

sub _new
{
   my $self = shift->SUPER::_new;
   ( $self->{size}, $self->{addr}, $self->{val}, my %args ) = @_;

   $self->{mask} = delete $args{mask} if exists $args{mask};

   $self->{loader} = $args{loader} if ref $args{loader};
   delete $args{loader};

   %args and die "Remaining args ".join(", ", keys %args);

   return $self;
}

sub instructions
{
   my $self = shift;
   my ( $jt, $jf ) = @_;

   my @values = ref $self->{val} ? @{ $self->{val} } : ( $self->{val} );

   my @jumps;
   foreach ( reverse @values ) {
      unshift @jumps, "JEQ $_, $jt, $jf";
      $jt++;
      $jf = 0;
   }

   return (
      ($self->{loader}||$self)->loadA_indir( $self->{size}, $self->{addr} ),
      ( exists $self->{mask} ? "AND $self->{mask}" : () ),
      @jumps,
   );
}

1;

package PktDump::Filter::Combined;
use base qw( PktDump::Filter );

sub _part { my $self = shift; push @{ $self->{parts} }, @_ }

package PktDump::Filter::All;
use base qw( PktDump::Filter::Combined );

sub instructions
{
   my $self = shift;
   my ( $jt, $jf ) = @_;

   my @instr = @{ $self->{instr} || [] };

   if( $jt > 0 ) {
      push @instr, "JA $jt";
   }

   $jt = 0;
   $jf += @instr;

   foreach my $part ( reverse @{ $self->{parts} } ) {
      my @parti = $part->instructions( $jt, $jf );
      unshift @instr, @parti;

      $jt = 0;
      $jf += @parti;
   }

   return @instr;
}

package PktDump::Filter::Any;
use base qw( PktDump::Filter::Combined );

sub instructions
{
   my $self = shift;
   my ( $jt, $jf ) = @_;

   $self->{instr} and die "TODO: Can't yet handle Any filter with epilogue\n";
   my @instr;

   foreach my $part ( reverse @{ $self->{parts} } ) {
      my @parti = $part->instructions( $jt, $jf );
      unshift @instr, @parti;

      $jt += @parti;
      $jf = 0;
   }

   return @instr;
}

package PktDump::Filter::Not;
use base qw( PktDump::Filter );

sub _new
{
   my $self = shift->SUPER::_new;
   ( $self->{inner} ) = @_;
   return $self;
}

sub instructions
{
   my $self = shift;
   my ( $jt, $jf ) = @_;
   return $self->{inner}->instructions( $jf, $jt );
}

package PktDump::Filter::Layer;
use base qw( PktDump::Filter );

use Linux::SocketFilter qw( SKF_NET_OFF );

use Carp;

sub extend
{
   my $self = shift;
   my ( $subfilter, %args ) = @_;

   my $offset    = $args{offset};
   my $genoffset = $args{genoffset};

   $offset eq "A" or croak "TODO: Can only handle offset at A currently";

   return [ PktDump::Filter->all(
         $self,
         $subfilter,
      ),
      genoffset => sub {
         $self->_setup_offset if !$self->{offset};

         # TODO: This is a bit of a hack
         my $offset = $self->{offset};
         $offset = SKF_NET_OFF if $offset eq "NET";

         $genoffset->(),
         "ADD $offset",
      },
      offset => "A"
   ];
}

sub matcher
{
   my $self = shift;
   my ( $filter, %args ) = @_;

   exists $args{offset} or croak "Need matcher offset";
   my $offset = delete $args{offset};

   my $genoffset = delete $args{genoffset};

   %args and croak "Other args found: " . join "|", keys %args;

   push @{ $self->{matchers} }, { 
      filter    => $filter,
      offset    => $offset,
      genoffset => $genoffset,
   };
}

sub instructions
{
   my $self = shift;
   my ( $jt, $jf ) = @_;

   my @instr;

   foreach my $matcher ( reverse @{ $self->{matchers} } ) {
      my $filter = $matcher->{filter};

      my @parti;

      if( $self->{enable_genoffset} ) {
         push @parti, $matcher->{genoffset}->() if $matcher->{genoffset};

         my $offset = $matcher->{offset};

         push @parti,
            $offset eq "X" ? "STX $self->{offset}" :
            $offset eq "A" ? "ST $self->{offset}" :
            die "Not sure how to handle offset $offset\n";
      }

      my $this_jt = $jt;

      if( @parti ) {
         if( $jt > 0 ) {
            push @parti, "JA $jt";
            $this_jt = 0;
         }

         $jf += @parti;
      }

      unshift @parti, $filter->instructions( $this_jt, $jf );

      unshift @instr, @parti;

      $jt += @parti;
      $jf = 0;
   }

   return @instr;
}

sub load_offset
{
   carp "load_offset";
   my $self = shift;
   my ( $r ) = @_;

   $r eq "A" or $r eq "X" or
      die "Expected load offset into A or X; found $r\n";

   return $r eq "X"
      ? "LDX $self->{offset}"
      : "LD $self->{offset}";
}

# These are static
my $scratch_num = 0;

sub _setup_offset
{
   my $self = shift;

   # Some easy cases
   NET: {
      $_->{offset} eq "NET" or last NET for @{ $self->{matchers} };

      $self->{offset} = "NET";
      return;
   }

   $self->{enable_genoffset} = 1;
   $self->{offset} = "M[$scratch_num]";
   $scratch_num++;
}

sub loadA_indir
{
   my $self = shift;
   my ( $size, $offset ) = @_;

   $self->_setup_offset if !$self->{offset};

   $offset =~ m/^-?\d+$/ or die "Expected immediate offset; found $offset\n";

   if( $self->{offset} =~ m/^-?\d+$/ ) {
      $offset += $self->{offset};
      return "LD $size\[$offset]";
   }

   if( $self->{offset} eq "NET" ) {
      return "LD $size\[NET+$offset]";
   }

   return (
      "LDX $self->{offset}",
      "LD $size\[X+$offset]",
   );
}

1;
